/*
 * DFT Spectrum analyzer
 */

#include "lcd.h"
#include "usart.h"
#include "pff.h"
#include "ffft.h"

#include <stdlib.h>
#include <string.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#define BIT8 0xFF
#define BIT7 0x01
#define BIT6 0x02
#define BIT5 0x03
#define BIT4 0x04
#define BIT3 0x05
#define BIT2 0x06
#define BIT1 0x07
#define BIT0 0xFE

//ideal 20 - 30
#define DELAY 100
#define START 0x80
#define NEXT_LINE 0x40

#define N 16

int16_t capture[FFT_N];			/* Wave captureing buffer */
complex_t bfly_buff[FFT_N];		/* FFT buffer */
uint16_t spektrum[FFT_N/2];		/* Spectrum output buffer */

int count = 0;

FATFS fs;					// sistemul de fisiere

void IO_init(void)
{
	DDRD &= ~((1 << PD6));	
	PORTD |= (1 << PD6);
}

/*---------------------------------------------------------------------------*/
/* Player audio                                                              */
/*---------------------------------------------------------------------------*/

/*
 * Four-Character Code - folosit pentru a indentifica formate de date
 */
#define FCC(c1, c2, c3, c4) \
	(((DWORD)(c4) << 24) + \
	 ((DWORD)(c3) << 16) + \
	 (( WORD)(c2) <<  8) + \
	 (( BYTE)(c1) <<  0))


	 uint8_t	buf[2][256];	// wave output buffers (double buffering)
const	 uint16_t	buf_size = 256;	// front and back buffer sizes
volatile uint8_t	buf_front = 0;	// front buffer index (current buffer used)
volatile uint8_t	buf_pos = 0;	// current buffer position
volatile uint8_t	buf_sync = 0;

#define BUF_FRONT	(buf[buf_front])
#define BUF_BACK	(buf[1 - buf_front])


ISR(TIMER0_COMPA_vect)
{
	OCR1A = BUF_FRONT[buf_pos++];

	if (count < FFT_N) {
		capture[count++] = BUF_FRONT[buf_pos];
	}

	// swap buffers when end is reached (end is 256 <=> overflow to 0)
	if(buf_pos == 0)
		buf_front = 1 - buf_front;
}

void timer0_start(void)
{
	// interrupt on compare A
	TIMSK0 |= (1 << OCIE0A);
	// CTC, top OCRA
	TCCR0B |= (0 << WGM02);
	TCCR0A |= (1 << WGM01) | (0 << WGM00);
	// prescaler 8
	TCCR0B |= (2 << CS00);
}

void timer0_stop(void)
{
	TCCR0B = 0;
	TCCR0A = 0;
	TIMSK0 = 0;
	OCR0A = 0;
	TCNT0 = 0;
}

void timer1_start(void)
{
	// 8-bit FastPWM
	TCCR1B |= (0 << WGM13) | (1 << WGM12);
	TCCR1A |= (0 << WGM11) | (1 << WGM10);
	// channel A inverted
	TCCR1A |= (3 << COM1A0);
	// prescaler 1
	TCCR1B |= (1 << CS10);
}

void timer1_stop(void)
{
	TCCR1B = 0;
	TCCR1A = 0;
	OCR1A = 0;
	TCNT1 = 0;
}

BOOL continue_play()
{	
	return TRUE;
}

/*
 * Incarca header-ul unui fisier WAVE
 *
 * @return DWORD
 * 	0 => format invalid
 * 	1 => eroare I/O
 * 	>1 => numarul de sample-uri
 */
DWORD load_header(void)
{
	DWORD size;
	WORD ret;

	// citeste header-ul (12 octeti)
	if(pf_read(BUF_FRONT, 12, &ret))
		return 1;

	if(ret != 12 || LD_DWORD(BUF_FRONT + 8) != FCC('W','A','V','E'))
		return 0;

	for(;;)
	{
		// citeste chunk ID si size
		pf_read(BUF_FRONT, 8, &ret);
		if(ret != 8)
			return 0;

		size = LD_DWORD(&BUF_FRONT[4]);

		// verifica FCC
		switch(LD_DWORD(&BUF_FRONT[0]))
		{
			// 'fmt ' chunk
			case FCC('f','m','t',' '):
				// verifica size
				if(size > 100 || size < 16) return 0;

				// citeste continutul
				pf_read(BUF_FRONT, size, &ret);
				// verifica codificarea
				if(ret != size || BUF_FRONT[0] != 1) return 0;
				// verifica numarul de canale
				if(BUF_FRONT[2] != 1 && BUF_FRONT[2] != 2) return 0;
				// verifica rezolutia
				if(BUF_FRONT[14] != 8 && BUF_FRONT[14] != 16) return 0;

				// seteaza sampling rate-ul
				OCR0A = (BYTE)(F_CPU / 8 / LD_WORD(&BUF_FRONT[4])) - 1;
				break;

			// 'data' chunk => incepe redarea
			case FCC('d','a','t','a'):
				return size;

			// 'LIST' chunk => skip
			case FCC('L','I','S','T'):
			// 'fact' chunk => skip
			case FCC('f','a','c','t'):
				pf_lseek(fs.fptr + size);
				break;

			// chunk necunoscut => eroare
			default:
				return 0;
		}
	}

	return 0;
}

/**
 * @addr - numarul coloanei unde pun o valoare
 * @state - valoarea intre 0 si 16 ce trebuie pusa
 * 			la coloana @addr
 */
void displayFrequencies(uint8_t addr, int state)
{
	if (state <= 8) {
		// linie jos
		uint8_t where = NEXT_LINE | addr;

		LCD_putCharAt(addr, BIT0);

		switch(state)
		{
			case 0:
				state++;
				LCD_putCharAt(where, BIT0);
				break;

			case 1:
				state++;
				LCD_putCharAt(where, BIT1);
				break;

			case 2:
				state++;
				LCD_putCharAt(where, BIT2);
				break;

			case 3:
				state++;
				LCD_putCharAt(where, BIT3);
				break;

			case 4:
				state++;
				LCD_putCharAt(where, BIT4);
				break;

			case 5:
				state++;
				LCD_putCharAt(where, BIT5);
				break;

			case 6:
				state++;
				LCD_putCharAt(where, BIT6);
				break;

			case 7:
				state++;
				LCD_putCharAt(where, BIT7);
				break;

			case 8:
				state++;
				LCD_putCharAt(where, BIT8);
				break;
		}
	} else {

		// linie sus
		LCD_putCharAt(NEXT_LINE | addr, BIT8);

		switch(state)
		{
			case 9:
				state++;
				LCD_putCharAt(addr, BIT1);
				break;

			case 10:
				state++;
				LCD_putCharAt(addr, BIT2);
				break;

			case 11:
				state++;
				LCD_putCharAt(addr, BIT3);
				break;

			case 12:
				state++;
				LCD_putCharAt(addr, BIT4);
				break;

			case 13:
				state++;
				LCD_putCharAt(addr, BIT5);
				break;

			case 14:
				state++;
				LCD_putCharAt(addr, BIT6);
				break;

			case 15:
				state++;
				LCD_putCharAt(addr, BIT7);
				break;

			case 16:
				state = 0;
				LCD_putCharAt(addr, BIT8);
				break;
		}
	}
}

// dummy values for testing animation
void demoFFT() {
	
	int i, value;

	for (i = 0; i < N; ++i) {
   		value = rand() % 17;
    	displayFrequencies(i, value);
    }
}

/*
 * Functie care reda un fisier audio
 *
 * path - calea absoluta a fisierului
 *
 * @return UINT
 *	FR_OK daca a rulat cu succes fisierul
 */
UINT play(const char *path)
{
	FRESULT ret;

	if((ret = pf_open(path)) == FR_OK)
	{
		WORD bytes_read;

		// incarca header-ul fisierului
		DWORD current_size = load_header();
		if(current_size < buf_size)
			return FR_NO_FILE;

		// align to sector boundary
		ret = pf_lseek((fs.fptr + 511) & ~511);
		if(ret != FR_OK)
			return ret;

		// fill front buffer
		ret = pf_read(BUF_FRONT, buf_size, &bytes_read);
		if(ret != FR_OK)
			return ret;
		if(bytes_read < buf_size)
			return ret;

		// reset front buffer index
		buf_pos = 0;

		// start output
		timer0_start();
		timer1_start();
		DDRD |= (1 << PD5);

		// redare muzica
		while(continue_play())
		{
			uint8_t old_buf_front = buf_front;
			
			// fill back buffer
			ret = pf_read(BUF_BACK, buf_size, &bytes_read);
			if(ret != FR_OK)
				break;
			if(bytes_read < buf_size)
				break;

			// wait for buffer swap
			while(old_buf_front == buf_front) ;
		}

		// stop output
		DDRD &= ~(1 << PD5);
		timer1_stop();
		timer0_stop();
	}

	return ret;
}


/*---------------------------------------------------------------------------*/
/* Ceas                                                                      */
/*---------------------------------------------------------------------------*/

volatile uint8_t hours = 0;
volatile uint8_t minutes = 0;
volatile uint8_t seconds = 0;


ISR(TIMER2_COMPA_vect)
{
	static uint16_t miliseconds = 0;

	// 125 best
	if(miliseconds++ != 125)
		return;
	miliseconds = 0;

	if(seconds++ == 60)
	{
		seconds = 0;

		if(minutes++ == 60)
		{
			minutes = 0;

			if(hours++ == 24)
				hours = 0;
		}
	}

	fft_input(capture, bfly_buff);
	fft_execute(bfly_buff);
	fft_output(bfly_buff, spektrum);
	
	count = 0;
	int i;

	for (i = 0; i < 16; ++i) {
		if (i == 15){
			spektrum[2*i + 1] *= 1.5f;
			displayFrequencies(i, spektrum[2*i + 1] + 1);
		} else {
			spektrum[2*i + 2] *= 1.5f;
			displayFrequencies(i, spektrum[2*i + 2] + 1);
		}
	}
}

void timer2_init(void)
{
	// set compare at each milisecond
	OCR2A = 124;
	// interrupt on compare A
	TIMSK2 |= (1 << OCIE2A);
	// CTC, top OCRA
	TCCR2B |= (0 << WGM22);
	TCCR2A |= (1 << WGM21) | (0 << WGM20);
	// prescaler 128
	TCCR2B |= (5 << CS20);
}

/*---------------------------------------------------------------------------*/
/* Task 2 and 3                                                              */
/*---------------------------------------------------------------------------*/

#define MUSIC "music"


/*
 * Cauta si returneaza al n-lea fisier cu extensia ".WAV" din folder-ul primit
 */
void get_music(int n, const char *folder, char *filename)
{
	filename[0] = '\0';

	DIR dir;
	if(pf_opendir(&dir, folder) != FR_OK)
		return;

	FILINFO file;
	while(n >= 0 && pf_readdir(&dir, &file) == FR_OK &&
		file.fname != NULL && file.fname[0] != '\0')
	{
		if((file.fattrib & (AM_DIR | AM_HID)) == 0 &&
			strstr(file.fname, ".WAV") != NULL)
		{
			n--;
		}
	}

	strcpy(filename, file.fname);
}

void spectrumAnalyzer()
{
	static int track = 0;
	static char path[] = MUSIC"/01-SUP~1.WAV";
	char * const filename = &path[sizeof(MUSIC)];

	LCD_clear_top_line();

	DIR dir;
	if(pf_opendir(&dir, MUSIC) != FR_OK)
	{
		// no music found
		return;
	}

		filename[-1] = '\0';
		get_music(track, path, filename);
		filename[-1] = '/';

		if(filename[0] == '\0')
		{
			track = 0;

			filename[-1] = '\0';
			get_music(track, path, filename);
			filename[-1] = '/';
		}

		_delay_ms(50);
	
	if(!(PIND & (1 << PD6)))
	{
		play(path);
		track++;
	}
}

void createCustomChars()
{
	LCD_writeInstr(LCD_INSTR_CGRAM);

	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	
	LCD_writeData(0x00);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x1F);
	LCD_writeData(0x1F);
	
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x00);
	LCD_writeData(0x1F);
}

int main()
{
	srand(13);   // should only be called once

	IO_init();
    LCD_init();
	timer2_init();

	sei();

    createCustomChars();
    LCD_writeInstr(LCD_INSTR_clearDisplay);

    for(;;)
	{
		// mount filesystem
		if(pf_mount(&fs) != FR_OK)
		{
			// wait a while and retry
			_delay_ms(1000);
			continue;
		}

		for(;;)
		{
			spectrumAnalyzer();

			_delay_ms(200);
		}
	}

    return 0;
}

